/* Broadcom NetXtreme-C/E network driver.
 *
 * Copyright (c) 2022 Broadcom Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation.
 */

#include <linux/stddef.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#ifdef HAVE_KTLS
#include <net/tls.h>
#endif

#include "bnxt_compat.h"
#include "bnxt_hsi.h"
#include "bnxt.h"
#include "bnxt_hwrm.h"
#include "bnxt_mpc.h"
#include "bnxt_ktls.h"

void bnxt_alloc_ktls_info(struct bnxt *bp, u16 max_key_ctxs_alloc)
{
	struct bnxt_ktls_info *ktls = bp->ktls_info;

	if (!ktls) {
		ktls = kzalloc(sizeof(*ktls), GFP_KERNEL);
		if (!ktls)
			return;
		ktls->tkc.max_ctx = BNXT_MAX_TKC;
		ktls->rkc.max_ctx = BNXT_MAX_RKC;
		INIT_LIST_HEAD(&ktls->tkc.list);
		spin_lock_init(&ktls->tkc.lock);
		INIT_LIST_HEAD(&ktls->rkc.list);
		spin_lock_init(&ktls->rkc.lock);
		bp->ktls_info = ktls;
	}
	ktls->max_key_ctxs_alloc = max_key_ctxs_alloc;
}

void bnxt_free_ktls_info(struct bnxt *bp)
{
	struct bnxt_ktls_info *ktls = bp->ktls_info;
	struct bnxt_kid_info *kid, *tmp;
	struct bnxt_kctx *kctx;

	if (!ktls)
		return;

	/* Shutting down, no need to protect the lists. */
	kctx = &ktls->tkc;
	list_for_each_entry_safe(kid, tmp, &kctx->list, list) {
		list_del(&kid->list);
		kfree(kid);
	}
	kctx = &ktls->rkc;
	list_for_each_entry_safe(kid, tmp, &kctx->list, list) {
		list_del(&kid->list);
		kfree(kid);
	}
	kfree(ktls);
	bp->ktls_info = NULL;
}

#if defined(HAVE_KTLS) && IS_ENABLED(CONFIG_TLS_DEVICE)
void bnxt_hwrm_reserve_pf_key_ctxs(struct bnxt *bp,
				   struct hwrm_func_cfg_input *req)
{
	struct bnxt_ktls_info *ktls = bp->ktls_info;
	struct bnxt_hw_resc *hw_resc = &bp->hw_resc;
	u16 tx, rx;

	if (!ktls)
		return;

	tx = min(ktls->tkc.max_ctx, hw_resc->max_tx_key_ctxs);
	req->num_tx_key_ctxs = cpu_to_le16(tx);
	rx = min(ktls->rkc.max_ctx, hw_resc->max_rx_key_ctxs);
	req->num_rx_key_ctxs = cpu_to_le16(rx);
	req->enables |= cpu_to_le32(FUNC_CFG_REQ_ENABLES_TX_KEY_CTXS |
				    FUNC_CFG_REQ_ENABLES_RX_KEY_CTXS);
}

static int bnxt_key_ctx_store(struct bnxt *bp, __le32 *key_buf, u16 num,
			      bool contig, struct bnxt_kctx *kctx)
{
	struct bnxt_kid_info *kid;
	u16 i;

	for (i = 0; i < num; ) {
		kid = kzalloc(sizeof(*kid), GFP_KERNEL);
		if (!kid)
			return -ENOMEM;
		kid->start_id = le32_to_cpu(key_buf[i]);
		if (contig)
			kid->count = num;
		else
			kid->count = 1;
		bitmap_set(kid->ids, 0, kid->count);
		spin_lock(&kctx->lock);
		list_add_tail_rcu(&kid->list, &kctx->list);
		kctx->total_alloc += kid->count;
		spin_unlock(&kctx->lock);
		i += kid->count;
	}
	return 0;
}

static int bnxt_hwrm_key_ctx_alloc(struct bnxt *bp, u8 type, u16 num,
				   struct bnxt_kctx *kctx)
{
	struct bnxt_ktls_info *ktls = bp->ktls_info;
	struct hwrm_func_key_ctx_alloc_output *resp;
	struct hwrm_func_key_ctx_alloc_input *req;
	dma_addr_t mapping;
	__le32 *key_buf;
	bool contig;
	int rc;

	rc = hwrm_req_init(bp, req, HWRM_FUNC_KEY_CTX_ALLOC);
	if (rc)
		return rc;

	num = min(num, ktls->max_key_ctxs_alloc);
	if ((num + kctx->total_alloc) > kctx->max_ctx)
		return -ENOSPC;

	key_buf = hwrm_req_dma_slice(bp, req, num * 4, &mapping);
	if (!key_buf) {
		rc = -ENOMEM;
		goto key_alloc_exit;
	}
	resp = hwrm_req_hold(bp, req);

	req->key_ctx_type = type;
	req->num_key_ctxs = cpu_to_le16(num);
	req->dma_bufr_size_bytes = cpu_to_le32(num * 4);
	req->host_dma_addr = cpu_to_le64(mapping);

	rc = hwrm_req_send(bp, req);
	if (rc)
		goto key_alloc_exit;

	num = le16_to_cpu(resp->num_key_ctxs_allocated);
	contig =
		resp->flags & FUNC_KEY_CTX_ALLOC_RESP_FLAGS_KEY_CTXS_CONTIGUOUS;
	rc = bnxt_key_ctx_store(bp, key_buf, num, contig, kctx);

key_alloc_exit:
	hwrm_req_drop(bp, req);
	return rc;
}

static int bnxt_key_ctx_alloc(struct bnxt *bp, u8 type, u16 num)
{
	struct bnxt_ktls_info *ktls = bp->ktls_info;
	struct bnxt_kctx *kctx;

	switch (type) {
	case BNXT_KID_TX:
		kctx = &ktls->tkc;
		break;
	case BNXT_KID_RX:
		kctx = &ktls->rkc;
		break;
	default:
		return -EINVAL;
	}
	return bnxt_hwrm_key_ctx_alloc(bp, type, num, kctx);
}

static int bnxt_alloc_one_kctx(struct bnxt_kctx *kctx, u32 *id)
{
	struct bnxt_kid_info *kid;
	int rc = -ENOMEM;

	rcu_read_lock();
	list_for_each_entry_rcu(kid, &kctx->list, list) {
		u32 idx = 0;

		do {
			idx = find_next_bit(kid->ids, kid->count, idx);
			if (idx >= kid->count)
				break;
			if (test_and_clear_bit(idx, kid->ids)) {
				*id = kid->start_id + idx;
				rc = 0;
				goto alloc_done;
			}
		} while (1);
	}

alloc_done:
	rcu_read_unlock();
	return rc;
}

static void bnxt_free_one_kctx(struct bnxt_kctx *kctx, u32 id)
{
	struct bnxt_kid_info *kid;

	rcu_read_lock();
	list_for_each_entry_rcu(kid, &kctx->list, list) {
		if (id >= kid->start_id && id < kid->start_id + kid->count) {
			set_bit(id - kid->start_id, kid->ids);
			break;
		}
	}
	rcu_read_unlock();
}

struct bnxt_crypto_cmd_ctx {
	struct completion cmp;
	struct ce_cmpl ce_cmp;
};

static int bnxt_xmit_crypto_cmd(struct bnxt *bp, struct bnxt_tx_ring_info *txr,
				void *cmd, uint len, uint tmo)
{
	struct bnxt_crypto_cmd_ctx ctx = {
		.cmp = COMPLETION_INITIALIZER_ONSTACK(ctx.cmp),
	};
	unsigned long tmo_left;
	int rc;

	spin_lock(&txr->tx_lock);
	rc = bnxt_start_xmit_mpc(bp, txr, cmd, len, (unsigned long)&ctx);
	spin_unlock(&txr->tx_lock);
	if (rc)
		return rc;
	tmo_left = wait_for_completion_timeout(&ctx.cmp, msecs_to_jiffies(tmo));
	if (!tmo_left)
		return -ETIMEDOUT;
	if (CE_CMPL_STATUS(&ctx.ce_cmp) == CE_CMPL_STATUS_OK)
		return 0;
	return -EIO;
}

static int bnxt_crypto_add(struct bnxt *bp,
			   enum tls_offload_ctx_dir direction,
			   struct tls_crypto_info *crypto_info,
			   struct bnxt_ktls_offload_ctx_tx *kctx_tx)
{
	struct bnxt_mpc_info *mpc = bp->mpc_info;
	struct bnxt_tx_ring_info *txr;
	struct ce_add_cmd cmd = {0};
	u32 data;

	if (direction == TLS_OFFLOAD_CTX_DIR_TX)
		txr = &mpc->mpc_rings[BNXT_MPC_TCE_TYPE][0];
	else
		txr = &mpc->mpc_rings[BNXT_MPC_RCE_TYPE][0];

	cmd.ctx_kind = CE_ADD_CMD_CTX_KIND_CK_TX;
	data = CE_ADD_CMD_OPCODE_ADD | (kctx_tx->kid << CE_ADD_CMD_KID_SFT);
	switch (crypto_info->cipher_type) {
	case TLS_CIPHER_AES_GCM_128: {
		struct tls12_crypto_info_aes_gcm_128 *aes;

		aes = (void *)crypto_info;
		data |= CE_ADD_CMD_ALGORITHM_AES_GCM_128;
		if (crypto_info->version == TLS_1_3_VERSION)
			data |= CE_ADD_CMD_VERSION_TLS1_3;
		memcpy(cmd.session_key, aes->key, sizeof(aes->key));
		memcpy(cmd.addl_iv, aes->iv, sizeof(aes->iv));
		memcpy(cmd.record_seq_num, aes->rec_seq, sizeof(aes->rec_seq));
		break;
	}
	case TLS_CIPHER_AES_GCM_256: {
		struct tls12_crypto_info_aes_gcm_256 *aes;

		aes = (void *)crypto_info;
		data |= CE_ADD_CMD_ALGORITHM_AES_GCM_256;
		if (crypto_info->version == TLS_1_3_VERSION)
			data |= CE_ADD_CMD_VERSION_TLS1_3;
		memcpy(cmd.session_key, aes->key, sizeof(aes->key));
		memcpy(cmd.addl_iv, aes->iv, sizeof(aes->iv));
		memcpy(cmd.record_seq_num, aes->rec_seq, sizeof(aes->rec_seq));
		break;
	}
	}
	cmd.ver_algo_kid_opcode = cpu_to_le32(data);
	cmd.pkt_tcp_seq_num = cpu_to_le32(kctx_tx->tcp_seq_no);
	cmd.tls_header_tcp_seq_num = cmd.pkt_tcp_seq_num;
	return bnxt_xmit_crypto_cmd(bp, txr, &cmd, sizeof(cmd),
				    BNXT_MPC_TMO_MSECS);
}

static int bnxt_crypto_del(struct bnxt *bp,
			   enum tls_offload_ctx_dir direction,
			   struct bnxt_ktls_offload_ctx_tx *kctx_tx)
{
	struct bnxt_mpc_info *mpc = bp->mpc_info;
	struct bnxt_tx_ring_info *txr;
	struct ce_delete_cmd cmd = {0};
	u32 data;

	if (direction == TLS_OFFLOAD_CTX_DIR_TX) {
		txr = &mpc->mpc_rings[BNXT_MPC_TCE_TYPE][0];
		data = CE_DELETE_CMD_CTX_KIND_CK_TX;
	} else {
		txr = &mpc->mpc_rings[BNXT_MPC_RCE_TYPE][0];
		data = CE_DELETE_CMD_CTX_KIND_CK_RX;
	}

	data |= CE_DELETE_CMD_OPCODE_DEL |
		(kctx_tx->kid << CE_DELETE_CMD_KID_SFT);

	cmd.ctx_kind_kid_opcode = cpu_to_le32(data);
	return bnxt_xmit_crypto_cmd(bp, txr, &cmd, sizeof(cmd),
				    BNXT_MPC_TMO_MSECS);
}

static bool bnxt_ktls_cipher_supported(struct bnxt *bp,
				       struct tls_crypto_info *crypto_info)
{
	u16 type = crypto_info->cipher_type;
	u16 version = crypto_info->version;

	if ((type == TLS_CIPHER_AES_GCM_128 ||
	     type == TLS_CIPHER_AES_GCM_256) &&
	    (version == TLS_1_2_VERSION ||
	     version == TLS_1_3_VERSION))
		return true;
	return false;
}

static int bnxt_ktls_dev_add(struct net_device *dev, struct sock *sk,
			     enum tls_offload_ctx_dir direction,
			     struct tls_crypto_info *crypto_info,
			     u32 start_offload_tcp_sn)
{
	struct bnxt_ktls_offload_ctx_tx *kctx_tx;
	struct bnxt *bp = netdev_priv(dev);
	struct bnxt_ktls_info *ktls;
	int rc;

	BUILD_BUG_ON(sizeof(struct bnxt_ktls_offload_ctx_tx) >
		     TLS_OFFLOAD_CONTEXT_SIZE_TX);

	if (direction == TLS_OFFLOAD_CTX_DIR_RX)
		return -EOPNOTSUPP;

	if (!bnxt_ktls_cipher_supported(bp, crypto_info))
		return -EOPNOTSUPP;

	kctx_tx = tls_driver_ctx(sk, direction);
	ktls = bp->ktls_info;
	rc = bnxt_alloc_one_kctx(&ktls->tkc, &kctx_tx->kid);
	if (rc)
		return rc;
	kctx_tx->tcp_seq_no = start_offload_tcp_sn;
	rc = bnxt_crypto_add(bp, direction, crypto_info, kctx_tx);
	if (rc)
		bnxt_free_one_kctx(&ktls->tkc, kctx_tx->kid);
	else
		atomic64_inc(&ktls->ktls_tx_add_count);
	return rc;
}

static void bnxt_ktls_dev_del(struct net_device *dev,
			      struct tls_context *tls_ctx,
			      enum tls_offload_ctx_dir direction)
{
	struct bnxt_ktls_offload_ctx_tx *kctx_tx;
	struct bnxt *bp = netdev_priv(dev);
	struct bnxt_ktls_info *ktls;
	int rc;

	if (direction != TLS_OFFLOAD_CTX_DIR_TX)
		return;

	ktls = bp->ktls_info;
	kctx_tx = __tls_driver_ctx(tls_ctx, direction);
	rc = bnxt_crypto_del(bp, direction, kctx_tx);
	if (!rc) {
		bnxt_free_one_kctx(&ktls->tkc, kctx_tx->kid);
		atomic64_inc(&ktls->ktls_tx_del_count);
	}
}

static const struct tlsdev_ops bnxt_ktls_ops = {
	.tls_dev_add = bnxt_ktls_dev_add,
	.tls_dev_del = bnxt_ktls_dev_del,
};

int bnxt_ktls_init(struct bnxt *bp)
{
	struct bnxt_ktls_info *ktls = bp->ktls_info;
	struct bnxt_hw_resc *hw_resc = &bp->hw_resc;
	struct net_device *dev = bp->dev;
	int rc;

	if (!ktls)
		return 0;

	ktls->tkc.max_ctx = hw_resc->resv_tx_key_ctxs;
	ktls->rkc.max_ctx = hw_resc->resv_rx_key_ctxs;

	if (!ktls->tkc.max_ctx || !ktls->rkc.max_ctx)
		return 0;

	rc = bnxt_key_ctx_alloc(bp, BNXT_KID_TX, BNXT_KID_BATCH_SIZE);
	if (rc)
		return rc;

	dev->tlsdev_ops = &bnxt_ktls_ops;
	dev->hw_features |= NETIF_F_HW_TLS_TX;
	dev->features |= NETIF_F_HW_TLS_TX;
	return 0;
}

void bnxt_ktls_mpc_cmp(struct bnxt *bp, unsigned long handle,
		       struct bnxt_cmpl_entry cmpl[], u32 entries)
{
	struct bnxt_crypto_cmd_ctx *ctx;
	u32 len;

	if (!handle || entries != 1) {
		netdev_warn(bp->dev, "Invalid handle %lx or entries %d in %s()\n",
			    handle, entries, __func__);
		return;
	}
	ctx = (void *)handle;
	len = min_t(u32, cmpl[0].len, sizeof(ctx->ce_cmp));
	memcpy(&ctx->ce_cmp, cmpl[0].cmpl, len);
	complete(&ctx->cmp);
}

#else	/* HAVE_KTLS */

void bnxt_hwrm_reserve_pf_key_ctxs(struct bnxt *bp,
				   struct hwrm_func_cfg_input *req)
{
}

int bnxt_ktls_init(struct bnxt *bp)
{
	return 0;
}

void bnxt_ktls_mpc_cmp(struct bnxt *bp, unsigned long handle,
		       struct bnxt_cmpl_entry cmpl[], u32 entries)
{
}
#endif	/* HAVE_KTLS */
